import unittest
from pyqint import PyQInt, Molecule
from copy import deepcopy
import numpy as np
import os

class TestCGF(unittest.TestCase):

    def testFunctionsCGF(self):
        """
        Test getting amplitude from CGF
        """

        # construct integrator object
        integrator = PyQInt()

        # get compile info
        compile_info = integrator.get_compile_info()

        # build hydrogen molecule
        mol = Molecule()
        mol.add_atom('H', 0.0, 0.0, 0.0)
        cgfs, nuclei = mol.build_basis('sto3g')

        # test values at these coordinates
        coords = []
        for x in np.linspace(0, 10, 10):
            coords.append([x, x, x])

        # results
        ans = [6.2825e-01, 7.1229e-02, 6.8672e-03, 3.0000e-04, 3.7662e-06,
               1.3536e-08, 1.3927e-11, 4.1021e-15, 3.4591e-19, 8.3505e-24]

        # test for each coord
        amps = []
        for i,coord in enumerate(coords):
            amp = cgfs[0].get_amp(coord)
            amps.append(amp)

        np.testing.assert_almost_equal(amps, ans, 4)

    def testPlotGrid(self):
        """
        Test plotting of 1b2 molecular orbital of H2O
        """
        # coefficients
        coeff = [8.37612e-17, -2.73592e-16,  -0.713011, -1.8627e-17, 9.53496e-17, -0.379323,  0.379323]

        # construct integrator object
        integrator = PyQInt()

        # build water molecule
        mol = Molecule('H2O')
        mol.add_atom('O', 0.0, 0.0, 0.0)
        mol.add_atom('H', 0.7570, 0.5860, 0.0)
        mol.add_atom('H', -0.7570, 0.5860, 0.0)
        cgfs, nuclei = mol.build_basis('sto3g')

        # build grid
        x = np.linspace(-2, 2, 50)
        y = np.linspace(-2, 2, 50)
        xx, yy = np.meshgrid(x,y)
        zz = np.zeros(len(x) * len(y))
        grid = np.vstack([xx.flatten(), yy.flatten(), zz]).reshape(3,-1).T
        res = integrator.plot_wavefunction(grid, coeff, cgfs).reshape((len(y), len(x)))

        ans = np.load(os.path.join(os.path.dirname(__file__), 'results', 'h2o_orb_1b2.npy'))
        np.testing.assert_almost_equal(res, ans, 6)

if __name__ == '__main__':
    unittest.main()



import os
from .molecule import Molecule 

class MoleculeBuilder:
    """
    Class that builds molecules from templates or from point group descriptions
    """
    def __init__(self):
        pass

    def from_name(self, molname):
        """
        Build molecule from molname
        """
        fname = os.path.join(os.path.dirname(__file__), 'molecules', molname.lower() + '.xyz')
        return self.from_file(fname)
       
    def from_file(self, path, molname=None):
        """
        Build molecule from file and return it
        """
        with open(path, 'r') as f:
            lines = f.readlines()
            
            nratoms = int(lines[0].strip())

            mol = Molecule(molname)
            for line in lines[2:2+nratoms]:
                pieces = line.split()
                mol.add_atom(pieces[0], float(pieces[1]), float(pieces[2]), float(pieces[3]), unit='angstrom')

            return mol

    def build_complex_td(self, r, at1, at2, unit='bohr'):
        mol = Molecule()
        mol.add_atom(at1,  0,  0,  0, unit=unit)
        mol.add_atom(at2,  1,  1,  1, unit=unit)
        mol.add_atom(at2,  1, -1, -1, unit=unit)
        mol.add_atom(at2, -1, -1,  1, unit=unit)
        mol.add_atom(at2, -1,  1,  1, unit=unit)

        return mol



import os
import tempfile
from pyqint import PyQInt
import numpy as np
import json
import subprocess
import shutil
from sys import platform

# try to import PyTessel but do not throw an error if it cannot be loaded
try:
    from pytessel import PyTessel
except ModuleNotFoundError:
    print('Cannot find module PyTessel')

try:
    from mendeleev import element
except ModuleNotFoundError:
    print('Cannot find module mendeleev')

try:
    from tqdm import tqdm
except ModuleNotFoundError:
    print('Cannot find module tqdm')

class BlenderRender:
    """
    This class leverages blender for rendering molecular orbitals
    """
    def __init__(self):
        self.executable = self.__find_blender()
        self.log = []
        print('******************************************************')
        print('WARNING: Blender rendering is an EXPERIMENTAL FEATURE.')
        print('******************************************************')
        if self.executable is None:
            raise Exception('Cannot find Blender executable')
        else:
            print('Found executable: %s' % self.executable)

    def render_molecular_orbitals(self, molecule, cgfs, orbc, outpath,
                                  mo_indices=None, sz=5, isovalue=0.03,
                                  prefix='MO', npts=100,
                                  negcol='E72F65', poscol='3F9EE7'):
        if mo_indices is None: # render all orbitals
            mo_indices = np.arange(0, len(orbc))

        # build a temporary folder
        tempdir = tempfile.mkdtemp()
        # store .xyz file
        xyzfile = os.path.join(tempdir, 'mol.xyz')
        self.__store_xyz(molecule, xyzfile)

        pbar = tqdm(mo_indices)
        for idx in pbar:
            # build isosurfaces
            pbar.set_description('Producing isosurfaces (#%i)' % (idx+1))
            plyfile = os.path.join(tempdir, '%s_%04i' % (prefix,idx+1))
            plypos, plyneg = self.__build_isosurface(plyfile, cgfs, orbc[:,idx], isovalue, sz, npts)

            # execute blender
            pbar.set_description('Producing molecular orbital (#%i)' % (idx+1))
            outfile = os.path.join(outpath, '%s_%04i.png' % (prefix,idx+1))
            logoutput = self.__run_blender(plypos, plyneg, xyzfile, outfile, tempdir, negcol, poscol)

            self.log.append("### START LOG: MOLECULAR ORBITAL %i ###" % (idx+1))
            for line in logoutput.splitlines():
                self.log.append(line.decode('utf-8'))
            self.log.append("### END LOG: MOLECULAR ORBITAL %i ###" % (idx+1))

        shutil.rmtree(tempdir)

        # store log in same path
        with open(os.path.join(outpath, 'renderlog.txt'), 'w') as f:
            for line in self.log:
                f.write(line + '\n')
            f.close()

    def get_executable(self):
        """
        Get the Blender executable
        """
        return self.executable

    def __find_blender(self):
        """
        Find the Blender executable
        """
        if platform == "linux" or platform == "linux2":
            ex = '/opt/blender-3.3.11-linux-x64/blender' # preferred version and path
            if os.path.exists(ex):
                return ex

            print('Cannot find proper Blender executable. For Linux, please install Blender LTS 3.3.11 in /opt/blender-3.3.11-linux-x64/.')
            print('Blender can be obtained via: https://ftp.nluug.nl/pub/graphics/blender/release/Blender3.3/blender-3.3.11-linux-x64.tar.xz')
            print('For more details on how to install Blender, please consult the instructions in the manual: https://pyqint.imc-tue.nl/')

            return None
        elif platform == 'win32':
            searchpath = os.path.join('C:\\','Program Files','Blender Foundation')
            name = 'blender.exe'
            results = []
            for root, dirs, files in os.walk(searchpath):
                if name in files:
                    results.append(os.path.join(root, name))

            for res in results:
                if '3.3' in res:
                    return res
        else:
            raise Exception('Your platform is not supported for Blender')

        return None

    def __run_blender(self, negfile, posfile, xyzfile, pngfile, cwd, negcol, poscol):
        # set path to xyz file
        blendpysrc = os.path.join(os.path.dirname(__file__), 'blender', 'blender_render_molecule.py')
        blendpydst = os.path.join(cwd, 'blender_render_molecule.py')
        shutil.copyfile(blendpysrc, blendpydst)

        manifest = {
            'mo_colors' : {
                'neg': negcol,
                'pos': poscol,
            },
            'xyzfile' : xyzfile,
            'mo_name' : 'isosurface',
            'mo_neg_path' : negfile,
            'mo_pos_path' : posfile,
            'png_output': pngfile,
            'bond_thickness': 0.2,
            'atom_radii' : {
                'H': 0.4,
                'N': 0.6,
                'C': 0.6,
                'O': 0.6,
            },
            'atom_colors' : {
                'H': 'FFFFFF',
                'N': '0000FF',
                'C': '000000',
                'O': 'DD0000',
            }
        }

        with open(os.path.join(cwd, 'manifest.json'), 'w') as f:
            f.write(json.dumps(manifest))

        # run blender
        out = subprocess.check_output(
            [self.executable, '-b', '-P', blendpydst],
            cwd=cwd
        )

        return out

    def __store_xyz(self, mol, filename):
        """
        Convert final result of GeometryOptimization class to an .xyz file
        which can be relayed to Blender for rendering

        Note that we have to convert Bohr units back to Angstrom as .xyz files
        are in units of Angstrom. Blender actually converts them from Angstrom back
        to Bohr, which looks like wasted effort but this way all .xyz files
        are consistent.
        """
        f = open(filename, 'w')
        f.write(str(len(mol.atoms)) + '\n')
        f.write('\n')

        angtobohr = 1.8897259886

        for a in mol.atoms:
            elname = element(a[0]).symbol
            f.write('%s  %12.6f  %12.6f  %12.6f\n' % (elname, a[1][0] / angtobohr,
                                                              a[1][1] / angtobohr,
                                                              a[1][2] / angtobohr))

        f.close()

    def __build_isosurface(self, filename, cgfs, coeff, isovalue, sz=5, npts=100):
        """
        Construct isosurfaces from PyQInt output
        """
        # generate some data
        isovalue = np.abs(isovalue)
        integrator = PyQInt()
        grid = integrator.build_rectgrid3d(-sz, sz, npts)
        scalarfield = np.reshape(integrator.plot_wavefunction(grid, coeff, cgfs), (npts, npts, npts))
        unitcell = np.diag(np.ones(3) * 2 * sz)

        pytessel = PyTessel()
        vertices, normals, indices = pytessel.marching_cubes(scalarfield.flatten(), scalarfield.shape, unitcell.flatten(), isovalue)
        posfile = filename + '_pos.ply'
        pytessel.write_ply(posfile, vertices, normals, indices)

        vertices, normals, indices = pytessel.marching_cubes(scalarfield.flatten(), scalarfield.shape, unitcell.flatten(), -isovalue)
        negfile = filename + '_neg.ply'
        pytessel.write_ply(negfile, vertices, normals, indices)

        return posfile, negfile



import unittest
from pyqint import Molecule, HF
from copy import deepcopy
import numpy as np

class TestHFDeriv(unittest.TestCase):

    def test_hartree_fock_forces_h2(self):
        """
        Test Hartree-Fock calculation on H2 using STO-3G basis set
        """
        for i in range(0,5):
            mol = Molecule()
            mol.add_atom('H', 0.0000, 0.0000, 0.30 + i * 0.50, unit='angstrom')
            mol.add_atom('H', 0.0000, 0.0000, -0.30 - i * 0.50, unit='angstrom')

            # calculate forces using analytical derivatives
            solver = HF()
            res = solver.rhf(mol, 'sto3g', calc_forces=True)

            # calculate forces using finite difference
            forces = calculate_forces_finite_difference(mol)

            np.testing.assert_almost_equal(res['forces'], forces, decimal=4)

    def test_hartree_fock_forces_h2o(self):
        """
        Test Hartree-Fock calculation on water using STO-3G basis set
        """
        mol = Molecule()
        mol.add_atom('O', 0.0, 0.1, 0.0)
        mol.add_atom('H', 0.7570, 0.5860, 0.0)
        mol.add_atom('H', -0.7570, 0.5860, 0.0)

        # calculate forces using analytical derivatives
        solver = HF()
        res = solver.rhf(mol, 'sto3g', calc_forces=True)

        # calculate forces using finite difference
        forces = calculate_forces_finite_difference(mol)

        np.testing.assert_almost_equal(res['forces'], forces, decimal=4)

    def test_hartree_fock_forces_ch4(self):
        """
        Test Hartree-Fock calculation on CH4 using STO-3G basis set

        Note that the CH4 molecule is slightly perturbed
        """
        mol = Molecule()
        mol.add_atom('C', 0.0,  0.1, 0.0, unit='angstrom')
        mol.add_atom('H', 0.1,  0.15, 1.0830098121, unit='angstrom')
        mol.add_atom('H', 0.0,  -1.0210714424, -0.3610032723, unit='angstrom')
        mol.add_atom('H', -0.8842738057,  0.5105357237, -0.3610032748, unit='angstrom')
        mol.add_atom('H', 0.8842738117,  0.5105357203, -0.3610032651, unit='angstrom')

        # calculate forces using analytical derivatives
        solver = HF()
        res = solver.rhf(mol, 'sto3g', calc_forces=True, tolerance=1e-12)

        # calculate forces using finite difference
        forces = calculate_forces_finite_difference(mol)

        np.testing.assert_almost_equal(res['forces'], forces, decimal=3)

    def test_hartree_fock_forces_co2(self):
        """
        Test Hartree-Fock calculation on CH4 using STO-3G basis set

        Note that the CO2 molecule is slightly perturbed
        """
        mol = Molecule()
        mol.add_atom('C', 0.0,  0.0, 0.0, unit='angstrom')
        mol.add_atom('O', 0.0,  0.0, 1.2879700928, unit='angstrom')
        mol.add_atom('O', 0.0,  0.0, -1.2879700928, unit='angstrom')

        # calculate forces using analytical derivatives
        solver = HF()
        res = solver.rhf(mol, 'sto3g', calc_forces=True)

        # calculate forces using finite difference
        forces = calculate_forces_finite_difference(mol)

        np.testing.assert_almost_equal(res['forces'], forces, decimal=3)

def perform_hf(mol):
    sol = HF().rhf(mol, 'sto3g', tolerance=1e-12)
    return sol

def calculate_forces_finite_difference(mol):
    """
    Calculates the forces on each of the atoms using a finite difference
    approach.
    """
    forces = np.zeros((len(mol.get_atoms()),3))

    sz = 1e-3

    for i in range(0, len(mol.get_atoms())): # loop over nuclei
        for j in range(0, 3): # loop over directions
            mol1 = deepcopy(mol)
            mol1.get_atoms()[i][1][j] -= sz / 2
            mol2 = deepcopy(mol)
            mol2.get_atoms()[i][1][j] += sz / 2

            energy1 = perform_hf(mol1)['energy']
            energy2 = perform_hf(mol2)['energy']

            forces[i,j] = (energy2 - energy1) / sz

    return forces

if __name__ == '__main__':
    unittest.main()



import unittest
from pyqint import PyQInt, gto, Molecule
import numpy as np

class TestOverlap(unittest.TestCase):

    def test_gto_overlap(self):
        """
        Test kinetic integrals for GTOs

        Sij = <gto_i | gto_j>
        """

        # construct integrator object
        integrator = PyQInt()

        # test GTO
        gto1 = gto(0.154329, [0.0, 0.0, 0.0], 3.425251, 0, 0, 0)
        gto2 = gto(0.535328, [0.0, 0.0, 0.0], 0.623914, 0, 0, 0)
        gto3 = gto(0.444635, [0.0, 0.0, 0.0], 0.168855, 0, 0, 0)
        overlap = integrator.overlap_gto(gto1, gto1)
        result = 0.31055691838264465
        np.testing.assert_almost_equal(overlap, result, 4)

    def test_cgf_overlap(self):
        """
        Test kinetic integrals for contracted Gaussians

        Sij = <cgf_i | cgf_j>
        """

        # construct integrator object
        integrator = PyQInt()

        # build hydrogen molecule
        mol = Molecule("H2")
        mol.add_atom('H', 0.0, 0.0, 0.0)
        mol.add_atom('H', 0.0, 0.0, 1.4)
        cgfs, nuclei = mol.build_basis('sto3g')

        S = np.zeros((2,2))
        S[0,0] = integrator.overlap(cgfs[0], cgfs[0])
        S[0,1] = S[1,0] = integrator.overlap(cgfs[0], cgfs[1])
        S[1,1] = integrator.overlap(cgfs[1], cgfs[1])

        S11 = 1.0
        S12 = 0.65931845
        np.testing.assert_almost_equal(S[0,0], S11, 4)
        np.testing.assert_almost_equal(S[1,1], S11, 4)
        np.testing.assert_almost_equal(S[0,1], S12, 4)

if __name__ == '__main__':
    unittest.main()



import unittest
from pyqint import Molecule, GeometryOptimization
import numpy as np

class TestHFMoleculeCharged(unittest.TestCase):

    def testHCO(self):
        """
        Test Hartree-Fock calculation for CO

        Data is compared to results obtained from Gaussian
        """
        mol = Molecule()
        mol.add_atom('O', -0.12770367,  -0.00000000,   1.23419284)
        mol.add_atom('C', -0.50625665,   0.00000000,  -1.09149431)
        mol.add_atom('H',  1.57882331,  -0.00000000,  -2.06681794)
        mol.set_charge(-1)

        results = GeometryOptimization().run(mol, 'sto3g')

        # check that energy matches
        self.assertEqual(results['data']['nelec'], 16)    
        np.testing.assert_almost_equal(results['energies'][-1], -111.523147758727, 5)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

import json
import os
import numpy as np
from .cgf import cgf
from .element import Element

class Molecule:
    """
    Molecule class
    """
    def __init__(self, _name='unknown'):
        self.__atoms = []
        self.__charges = []
        self.name = _name
        self.__charge = 0
        self.__nelec = None

    def __str__(self):
        res = "Molecule: %s\n" % self.name
        for atom in self.__atoms:
            res += " %s (%f,%f,%f)\n" % (atom[0], atom[1][0], atom[1][1], atom[1][2])

        return res

    def get_nelec(self):
        """
        Get the number of electrons
        """
        if self.__nelec == None:
            raise Exception('You need to use build_basis() or get_nuclei() before using this function.')
        return self.__nelec - self.__charge

    def get_atoms(self):
        return self.__atoms
    
    def get_charge(self):
        return self.__charge

    def set_charge(self, charge):
        """
        Set the charge of the molecule
        """
        self.__charge = charge

    def add_atom(self, atom, x, y, z, unit='bohr'):
        """
        Add an atom to the molecule
        """
        ang2bohr = 1.8897259886

        x = float(x)
        y = float(y)
        z = float(z)

        if unit == "bohr":
            self.__atoms.append([atom, np.array([x, y, z])])
        elif unit == "angstrom":
            self.__atoms.append([atom, np.array([x*ang2bohr, y*ang2bohr, z*ang2bohr])])
        else:
            raise RuntimeError("Invalid unit encountered: %s. Accepted units are 'bohr' and 'angstrom'." % unit)

        self.__charges.append(0)

    def build_basis(self, name):
        """
        Build a basis set from a label

        Returns list of CGFs and nuclei
        """
        basis_filename = os.path.join(os.path.dirname(__file__), 'basissets', '%s.json' % name)
        f = open(basis_filename, 'r')
        basis = json.load(f)
        f.close()
        
        self.__cgfs = []

        for aidx, atom in enumerate(self.__atoms):
            cgfs_template = basis[atom[0]]

            # store information about the nuclei
            self.__charges[aidx] = cgfs_template['atomic_number']

            for cgf_t in cgfs_template['cgfs']:
                # s-orbitals
                if cgf_t['type'] == 'S':
                    self.__cgfs.append(cgf(atom[1]))
                    for gto in cgf_t['gtos']:
                        self.__cgfs[-1].add_gto(gto['coeff'], gto['alpha'], 0, 0, 0)

                # p-orbitals
                if cgf_t['type'] == 'P':
                    self.__cgfs.append(cgf(atom[1]))
                    for gto in cgf_t['gtos']:
                        self.__cgfs[-1].add_gto(gto['coeff'], gto['alpha'], 1, 0, 0)
                    
                    self.__cgfs.append(cgf(atom[1]))
                    for gto in cgf_t['gtos']:
                        self.__cgfs[-1].add_gto(gto['coeff'], gto['alpha'], 0, 1, 0)
                    
                    self.__cgfs.append(cgf(atom[1]))
                    for gto in cgf_t['gtos']:
                        self.__cgfs[-1].add_gto(gto['coeff'], gto['alpha'], 0, 0, 1)

                # d-orbitals
                if cgf_t['type'] == 'D':
                    self.__cgfs.append(cgf(atom[1]))
                    for gto in cgf_t['gtos']:
                        self.__cgfs[-1].add_gto(gto['coeff'], gto['alpha'], 2, 0, 0)
                    
                    self.__cgfs.append(cgf(atom[1]))
                    for gto in cgf_t['gtos']:
                        self.__cgfs[-1].add_gto(gto['coeff'], gto['alpha'], 0, 2, 0)
                    
                    self.__cgfs.append(cgf(atom[1]))
                    for gto in cgf_t['gtos']:
                        self.__cgfs[-1].add_gto(gto['coeff'], gto['alpha'], 0, 0, 2)
                    
                    self.__cgfs.append(cgf(atom[1]))
                    for gto in cgf_t['gtos']:
                        self.__cgfs[-1].add_gto(gto['coeff'], gto['alpha'], 1, 1, 0)
                    
                    self.__cgfs.append(cgf(atom[1]))
                    for gto in cgf_t['gtos']:
                        self.__cgfs[-1].add_gto(gto['coeff'], gto['alpha'], 1, 0, 1)
                    
                    self.__cgfs.append(cgf(atom[1]))
                    for gto in cgf_t['gtos']:
                        self.__cgfs[-1].add_gto(gto['coeff'], gto['alpha'], 0, 1, 1)

        # build nuclei objects
        self.get_nuclei()
        self.__nelec = np.sum(self.__charges)

        return self.__cgfs, self.__nuclei
    
    def get_nuclei(self):
        """
        Get the nuclei as a packed array
        """
        el = Element()

        # reset list
        self.__nuclei = []

        for aidx, atom in enumerate(self.__atoms):

            # store information about the nuclei
            self.__charges[aidx] = getattr(el, atom[0])
            self.__nuclei.append([atom[1], self.__charges[aidx]])

        # populate number of electrons
        self.__nelec = np.sum(self.__charges)

        return self.__nuclei